package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.util.WaitForTaskExecutionsTask;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.ListSliceMembersTask;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import javafx.application.Platform;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.control.cell.CheckBoxListCell;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import javafx.util.Callback;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

/**
 */
public class UnshareSliceDialog extends VBox {
    private static final Logger LOG = LoggerFactory.getLogger(UnshareSliceDialog.class);

    @Nonnull private final Slice slice;
    @Nonnull private final ExperimentController experimentController;

    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final SfaModel sfaModel;
    @Nonnull private final HighLevelTaskFactory hltf;


    @FXML
    private ListView<SelectableUser> usersList;

    @FXML
    private HBox initializingBox;

    @FXML
    private CheckBox removeSshKeysCheckBox;

    private final Dialog<ButtonType> dialog;
    private final ButtonType unshareType = new ButtonType("Unshare", ButtonBar.ButtonData.OK_DONE);

    UnshareSliceDialog(@Nullable Window parentWindow, @Nonnull ExperimentController experimentController,
                       @Nonnull GeniUserProvider geniUserProvider, @Nonnull SfaModel sfaModel, @Nonnull HighLevelTaskFactory hltf) {
        if (experimentController.getExperiment().getSliceOrNull() == null) {
            throw new IllegalArgumentException("Can't create RenewSliceDialog if there experiment has slice == null");
        }

        this.experimentController = experimentController;
        this.geniUserProvider = geniUserProvider;
        this.sfaModel = sfaModel;
        this.hltf = hltf;
        this.slice = experimentController.getExperiment().getSliceOrNull();

        //initialize the Scene
        FXMLUtil.injectFXML(this);


        usersList.setCellFactory(new Callback<ListView<SelectableUser>, ListCell<SelectableUser>>() {
            @Override
            public ListCell<SelectableUser> call(ListView<SelectableUser> param) {
                return new CheckBoxListCell<SelectableUser>(SelectableUser::selectedProperty) {
                    @Override
                    public void updateItem(SelectableUser item, boolean empty) {
                        super.updateItem(item, empty);

                        if (item != null) {
                            setDisable(item.isFixed());
                        }
                    }
                };
            }
        });

        WaitForTaskExecutionsTask waitTask = new WaitForTaskExecutionsTask() {

            @Override
            public void onTaskExecutionsFinished() {
                Platform.runLater(() -> {
                    updateUsersList();

                    initializingBox.setVisible(false);
                });

            }
        };

        if (experimentController.getExperiment().getSliceOrNull().getUsers().isEmpty()) {
            ListSliceMembersTask listSliceMembersTask =
                    this.hltf.listSliceMembers(experimentController.getExperiment().getSliceOrNull());

            waitTask.addTaskExecution(this.hltf.submitTask(listSliceMembersTask));
        }

        new Thread(null, waitTask, "UnshareSlice-WaitForBTaskExecutionsTasks").start();

        final String sliceProject = slice.getSubAuthorityName();
        Server userAuthServer = geniUserProvider.getLoggedInGeniUser().getUserAuthorityServer();
        final String authName = userAuthServer == null ? null : userAuthServer.getName();

        //initialize dialog
        dialog = new Dialog<>();
        dialog.initOwner(parentWindow);
        dialog.setTitle("Unshare Experiment");
        dialog.setHeaderText("Unshare your experiment");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.AMBULANCE, Color.GREEN));
        dialog.getDialogPane().setContent(this);
        dialog.getDialogPane().getButtonTypes().setAll(unshareType, ButtonType.CANCEL);

        Node unshareButton = dialog.getDialogPane().lookupButton(unshareType);
        unshareButton.disableProperty().bind(selectedNewUsersCount.isEqualTo(0));

        dialog.resultProperty().addListener(observable -> {
            if (dialog.getResult() == unshareType) {
                List<GeniUrn> selectedUsers = usersList.getItems().stream()
                        .filter(SelectableUser::getSelected)
                        .map(SelectableUser::getUrn)
                        .collect(Collectors.toList());

                if (!selectedUsers.isEmpty())
                    experimentController.unshareWithUsers(selectedUsers, removeSshKeysCheckBox.isSelected());
                else
                    LOG.warn("Unexpectedly got an empty list of slice members to remove");
            }
        });

    }

    private void updateUsersList() {
        assert !experimentController.getExperiment().getSliceOrNull().getUsers().isEmpty();

        usersList.getItems().clear();

        final Slice slice = experimentController.getExperiment().getSliceOrNull();
        final Collection<GeniUrn> currentMembers = slice == null ? Collections.emptyList() : slice.getUsers();
        for (GeniUrn currentMember : currentMembers) {
            boolean owner = slice != null &&
                    slice.getCreator_urn() != null &&
                    currentMember.toString().equals(slice.getCreator_urn());
            usersList.getItems().addAll(new SelectableUser(currentMember, false, owner));
        }

        if (slice != null)
            LOG.debug("Slice owner: "+slice.getCreator_urn());

        usersList.getItems().sort(
                Comparator.comparing(su -> su.getUrn().getResourceName().toLowerCase())
        );
    }


    public void showDialog() {
        dialog.show();
    }

    private final IntegerProperty selectedNewUsersCount = new SimpleIntegerProperty(0);

    private final class SelectableUser {

        private final GeniUrn urn;
        private final BooleanProperty selected = new SimpleBooleanProperty(false);
        private final boolean fixed;


        public SelectableUser(GeniUrn urn, boolean selected, boolean fixed) {
            this.fixed = fixed;

            if (!fixed) {
                this.selected.addListener((observable, oldValue, newValue) -> {
                    if (oldValue != newValue) {
                        if (newValue) {
                            selectedNewUsersCount.set(selectedNewUsersCount.get() + 1);
                        } else {
                            selectedNewUsersCount.set(selectedNewUsersCount.get() - 1);
                        }
                    }
                });
            }

            this.selected.set(selected);
            this.urn = urn;
        }

        public boolean isFixed() {
            return fixed;
        }

        public boolean getSelected() {
            return selected.get();
        }

        public BooleanProperty selectedProperty() {
            return selected;
        }

        public void setSelected(boolean selected) {
            this.selected.set(selected);
        }

        public GeniUrn getUrn() {
            return urn;
        }

        @Override
        public String toString() {
            return urn.getResourceName();
        }
    }
}
